# pylint: disable=too-many-locals
from typing import Dict, List

import pytest
from pytest_test_utils import TmpDir

from gto.registry import GitRegistry

from .utils import check_obj

EXPECTED_REGISTRY_TAG_TAG_STATE = {
    "artifacts": {
        "rf": {
            "artifact": "rf",
            "versions": [
                {
                    "artifact": "rf",
                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                    "version": "v1.2.3",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:57",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Add artifacts",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v1.2.3",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/random-forest.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [
                        {
                            "priority": 3,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:57",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Registering artifact rf version v1.2.3",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "tag": "rf@v1.2.3",
                            "version": "v1.2.3",
                        }
                    ],
                    "deregistrations": [],
                    "stages": {
                        "production": {
                            "artifact": "rf",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v1.2.3",
                            "stage": "production",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:56:59",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage production to artifact rf version v1.2.3",
                                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                                    "tag": "rf#production#1",
                                    "version": "v1.2.3",
                                    "stage": "production",
                                },
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:57:02",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage production to artifact rf version v1.2.3",
                                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                                    "tag": "rf#production#4",
                                    "version": "v1.2.3",
                                    "stage": "production",
                                },
                            ],
                            "unassignments": [],
                        }
                    },
                },
                {
                    "artifact": "rf",
                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                    "version": "v1.2.4",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Update model",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "v1.2.4",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/random-forest.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [
                        {
                            "priority": 3,
                            "addition": True,
                            "artifact": "rf",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Registering artifact rf version v1.2.4",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "tag": "rf@v1.2.4",
                            "version": "v1.2.4",
                        }
                    ],
                    "deregistrations": [],
                    "stages": {
                        "staging": {
                            "artifact": "rf",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "v1.2.4",
                            "stage": "staging",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:57:00",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage staging to artifact rf version v1.2.4",
                                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                                    "tag": "rf#staging#2",
                                    "version": "v1.2.4",
                                    "stage": "staging",
                                }
                            ],
                            "unassignments": [],
                        },
                        "production": {
                            "artifact": "rf",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "v1.2.4",
                            "stage": "production",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "rf",
                                    "created_at": "2022-08-04T16:57:01",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage production to artifact rf version v1.2.4",
                                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                                    "tag": "rf#production#3",
                                    "version": "v1.2.4",
                                    "stage": "production",
                                }
                            ],
                            "unassignments": [],
                        },
                    },
                },
            ],
            "creations": [],
            "deprecations": [],
        },
        "nn": {
            "artifact": "nn",
            "versions": [
                {
                    "artifact": "nn",
                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                    "version": "v0.0.1",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "nn",
                            "created_at": "2022-08-04T16:56:57",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Add artifacts",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v0.0.1",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/neural-network.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [
                        {
                            "priority": 3,
                            "addition": True,
                            "artifact": "nn",
                            "created_at": "2022-08-04T16:56:58",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Registering artifact nn version v0.0.1",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "tag": "nn@v0.0.1",
                            "version": "v0.0.1",
                        }
                    ],
                    "deregistrations": [],
                    "stages": {
                        "staging": {
                            "artifact": "nn",
                            "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                            "version": "v0.0.1",
                            "stage": "staging",
                            "assignments": [
                                {
                                    "priority": 5,
                                    "addition": True,
                                    "artifact": "nn",
                                    "created_at": "2022-08-04T16:56:59",
                                    "author": "Alexander Guschin",
                                    "author_email": "1aguschin@gmail.com",
                                    "message": "Assigning stage staging to artifact nn version v0.0.1",
                                    "commit_hexsha": "89de382074d472f8e6b8fd654490183c3c0fb497",
                                    "tag": "nn#staging#1",
                                    "version": "v0.0.1",
                                    "stage": "staging",
                                }
                            ],
                            "unassignments": [],
                        }
                    },
                },
                {
                    "artifact": "nn",
                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                    "version": "04d79900801d9aa7ec726169706280a32a25d198",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "nn",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Update model",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "04d79900801d9aa7ec726169706280a32a25d198",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "model",
                                        "path": "models/neural-network.pkl",
                                        "virtual": False,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [],
                    "deregistrations": [],
                    "stages": {},
                },
            ],
            "creations": [],
            "deprecations": [],
        },
        "features": {
            "artifact": "features",
            "versions": [
                {
                    "artifact": "features",
                    "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                    "version": "04d79900801d9aa7ec726169706280a32a25d198",
                    "enrichments": [
                        {
                            "priority": 0,
                            "addition": True,
                            "artifact": "features",
                            "created_at": "2022-08-04T16:56:59",
                            "author": "Alexander Guschin",
                            "author_email": "1aguschin@gmail.com",
                            "message": "Update model",
                            "commit_hexsha": "04d79900801d9aa7ec726169706280a32a25d198",
                            "version": "04d79900801d9aa7ec726169706280a32a25d198",
                            "enrichments": [
                                {
                                    "source": "gto",
                                    "artifact": {
                                        "type": "dataset",
                                        "path": "datasets/features.csv",
                                        "virtual": True,
                                        "labels": [],
                                        "description": "",
                                        "custom": None,
                                    },
                                }
                            ],
                            "committer": "Alexander Guschin",
                            "committer_email": "1aguschin@gmail.com",
                        }
                    ],
                    "registrations": [],
                    "deregistrations": [],
                    "stages": {},
                }
            ],
            "creations": [],
            "deprecations": [],
        },
    }
}


def iter_over(sequence):
    if isinstance(sequence, dict):
        return list(sequence.values())
    if isinstance(sequence, list):
        return sequence
    raise NotImplementedError


@pytest.mark.usefixtures("showcase")
def test_registry_state_tag_tag(tmp_dir: TmpDir):
    with GitRegistry.from_url(tmp_dir) as reg:
        appeared_state = reg.get_state().dict()

    # TODO: update state
    exclude: Dict[str, List[str]] = {
        "commits": [],
        "versions": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "registrations",
            "deregistrations",
            "enrichments",
            "stages",
            "message",
            "version",
        ],
        "enrichments": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
            "committer",
            "committer_email",
        ],
        "registrations": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
        ],
        "deregistrations": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
        ],
        "stages": [
            "author",
            "author_email",
            "created_at",
            "commit_hexsha",
            "message",
            "version",
            "assignments",
            "unassignments",
        ],
    }

    expected_state = EXPECTED_REGISTRY_TAG_TAG_STATE
    for artifact in expected_state["artifacts"]:

        for appeared, expected in zip(
            iter_over(appeared_state["artifacts"][artifact]["versions"]),
            iter_over(expected_state["artifacts"][artifact]["versions"]),
        ):
            check_obj(appeared, expected, exclude["versions"])

            for key in ["enrichments", "registrations", "deregistrations"]:
                for a, e in zip(
                    iter_over(appeared[key]),
                    iter_over(expected[key]),
                ):
                    check_obj(a, e, exclude[key])

            for key in appeared["stages"]:
                check_obj(
                    appeared["stages"][key], expected["stages"][key], exclude["stages"]
                )



# pylint: disable=redefined-outer-name
import sys
from time import sleep
from typing import Tuple

import pytest
from click.testing import Result
from pytest_test_utils import TmpDir
from scmrepo.git import Git
from typer.testing import CliRunner

import gto
from gto.cli import app
from gto.config import CONFIG_FILE_NAME


class Runner:
    def __init__(self):
        self._runner = CliRunner(mix_stderr=False)

    def invoke(self, *args, **kwargs) -> Result:
        return self._runner.invoke(app, *args, **kwargs)


@pytest.fixture
def runner() -> Runner:
    return Runner()


@pytest.fixture
def scm(tmp_dir: TmpDir) -> Git:
    scm_instance = Git.init(tmp_dir)
    try:
        yield scm_instance
    finally:
        scm_instance.close()


@pytest.fixture(autouse=True)
def set_recursion_limit() -> None:
    sys.setrecursionlimit(2000)


@pytest.fixture
def empty_git_repo(scm: Git) -> str:
    return scm.root_dir


@pytest.fixture
def repo_with_commit(tmp_dir: TmpDir, scm: Git) -> str:
    tmp_dir.gen(
        "some-tracked-file",
        "some-random-text",
    )
    scm.add("some-tracked-file")
    scm.commit("Initial commit")
    return str(tmp_dir)


@pytest.fixture
def init_showcase_semver(tmp_dir: TmpDir, request: pytest.FixtureRequest):
    request.getfixturevalue("scm")
    tmp_dir.gen(CONFIG_FILE_NAME, "")


artifacts_yaml = """
rf:
  type: model
  path: models/random-forest.pkl
  virtual: false
nn:
  type: model
  path: models/neural-network.pkl
  virtual: false
features:
  type: dataset
  path: datasets/features.csv
"""


@pytest.fixture
def showcase(
    tmp_dir: TmpDir, scm: Git, request: pytest.FixtureRequest
) -> Tuple[str, str]:  # pylint: disable=too-many-locals, too-many-statements
    request.getfixturevalue("init_showcase_semver")
    tmp_dir.gen(
        {
            "models/random-forest.pkl": "1st version",
            "models/neural-network.pkl": "1st version",
        }
    )
    scm.add(["models"])
    scm.commit("Create models")

    tmp_dir.gen("artifacts.yaml", artifacts_yaml)
    scm.add(["artifacts.yaml"])
    scm.commit("Add artifacts")
    first_commit = scm.get_rev()

    nn_vname = "v0.0.1"
    rf_vname = "v1.2.3"
    gto.api.register(tmp_dir, "rf", "HEAD", rf_vname)
    gto.api.register(tmp_dir, "nn", "HEAD")
    sleep(1)

    tmp_dir.gen("models/random-forest.pkl", "2nd version")

    scm.commit("Update model")
    second_commit = scm.get_rev()

    # bump version automatically
    gto.api.register(tmp_dir, "rf", "HEAD")

    gto.api.assign(tmp_dir, "nn", "staging", version=nn_vname)
    gto.api.assign(tmp_dir, "rf", "production", version=rf_vname)
    sleep(1)  # this is needed to ensure right order of assignments in later checks
    # the problem is git tags doesn't have miliseconds precision, so we need to wait a bit
    gto.api.assign(tmp_dir, "rf", "staging", ref="HEAD")
    sleep(1)
    gto.api.assign(tmp_dir, "rf", "production", ref=scm.get_rev())
    sleep(1)
    gto.api.assign(tmp_dir, "rf", "production", version=rf_vname, force=True)
    return first_commit, second_commit



import pytest
from pytest_test_utils import TmpDir
from scmrepo.git import Git
from typer.testing import CliRunner

from gto.api import assign, get_stages, register
from gto.cli import app
from gto.config import CONFIG_FILE_NAME
from gto.exceptions import InvalidVersion, UnknownStage, ValidationError
from gto.index import RepoIndexManager
from gto.registry import GitRegistry

CONFIG_CONTENT = """
types: [model, dataset]
stages: [dev, prod]
"""

PROHIBIT_CONFIG_CONTENT = """
types: []
stages: []
"""

ALLOWED_STRING = "model"
DISALLOWED_STRING = "###"


@pytest.fixture(name="init_repo")
def _init_repo(tmp_dir: TmpDir, scm: Git) -> TmpDir:
    tmp_dir.gen(CONFIG_FILE_NAME, CONFIG_CONTENT)
    scm.add([CONFIG_FILE_NAME])
    scm.commit("Initial commit")
    return tmp_dir


def test_config_load_index(init_repo: TmpDir):
    with RepoIndexManager.from_url(init_repo) as index:
        assert index.config.TYPES == ["model", "dataset"]


def test_config_load_registry(init_repo: TmpDir):
    with GitRegistry.from_url(init_repo) as reg:
        assert reg.config.TYPES == ["model", "dataset"]


def test_stages(init_repo: TmpDir):
    assert get_stages(init_repo) == ["dev", "prod"]
    assert get_stages(init_repo, allowed=True) == ["dev", "prod"]
    assert get_stages(init_repo, used=True) == []


def test_register_incorrect_name(init_repo: TmpDir):
    with pytest.raises(ValidationError):
        register(init_repo, DISALLOWED_STRING, ref="HEAD")


def test_register_incorrect_version(init_repo: TmpDir):
    with pytest.raises(InvalidVersion):
        register(init_repo, "model", ref="HEAD", version="###")


def test_assign_incorrect_name(init_repo: TmpDir):
    with pytest.raises(ValidationError):
        assign(init_repo, DISALLOWED_STRING, ref="HEAD", stage="dev")


def test_assign_incorrect_stage(init_repo: TmpDir):
    with pytest.raises(ValidationError):
        assign(init_repo, ALLOWED_STRING, ref="HEAD", stage=DISALLOWED_STRING)


@pytest.mark.usefixtures("scm")
def test_config_is_not_needed(tmp_dir: TmpDir):
    tmp_dir.gen(
        CONFIG_FILE_NAME,
        "WRONG_CONFIG",
    )
    runner = CliRunner()
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0


@pytest.fixture(name="init_repo_prohibit")
def _init_repo_prohibit(tmp_dir: TmpDir, scm: Git) -> TmpDir:
    tmp_dir.gen(CONFIG_FILE_NAME, PROHIBIT_CONFIG_CONTENT)
    scm.add(CONFIG_FILE_NAME)
    scm.commit("Initial commit")
    return tmp_dir


def test_prohibit_config_assign_incorrect_stage(init_repo_prohibit: TmpDir):
    with pytest.raises(UnknownStage):
        assign(init_repo_prohibit, ALLOWED_STRING, ref="HEAD", stage="dev")



from typing import Sequence

import pytest
from pytest_test_utils import TmpDir
from scmrepo.git import Git

from gto import CONFIG
from gto.index import (
    Artifact,
    RepoIndexManager,
    check_if_path_exists,
    find_repeated_path,
)


@pytest.fixture(name="index")
def _index(empty_git_repo: str) -> RepoIndexManager:
    with RepoIndexManager.from_url(empty_git_repo) as index:
        yield index


def test_git_index_add_virtual(tmp_dir: TmpDir, scm: Git, index: RepoIndexManager):
    index.add(
        "nn",
        type="tt",
        path="pp",
        must_exist=False,
        allow_same_path=False,
        labels=[],
        description="",
        custom=None,
        update=False,
        commit=False,
        commit_message=None,
    )

    with RepoIndexManager.from_url(tmp_dir) as new_index:
        assert isinstance(new_index, RepoIndexManager)
        index_value = new_index.get_index()
        assert index_value.state["nn"] == Artifact(path="pp", type="tt", virtual=True)

        scm.add(CONFIG.INDEX)
        scm.commit("add index")
        rev = scm.get_rev()

        assert new_index.get_history()[rev].state == index_value.state


def test_git_index_remove_virtual(tmp_dir: TmpDir, index: RepoIndexManager):
    index.add(
        "aa",
        "aa",
        "aa",
        must_exist=False,
        allow_same_path=False,
        labels=[],
        description="",
        custom=None,
        update=True,
        commit=False,
        commit_message=None,
    )

    with RepoIndexManager.from_url(tmp_dir) as new_index:
        assert isinstance(new_index, RepoIndexManager)

        new_index.remove("aa")
        index_value = new_index.get_index()
        assert index_value.state == {}


@pytest.mark.parametrize(
    "path, paths",
    [
        ("models/m1.txt", ["models/m1.txt"]),
        ("models/m1.txt", ["models"]),
        ("models", ["models/m1.txt"]),
        ("models", ["models/"]),
        ("models/m1.txt/", ["models/m1.txt"]),
    ],
)
def test_check_path_found(path: str, paths: Sequence[str]):
    assert find_repeated_path(path, paths) is not None


@pytest.mark.parametrize(
    "path, paths",
    [("models/m1.txt", ["models/m2.txt"]), ("models/a/m1.txt", ["models/m1.txt"])],
)
def test_check_path_not_found(path: str, paths: Sequence[str]):
    assert find_repeated_path(path, paths) is None


def test_check_existence_repo(tmp_dir: TmpDir, scm: Git):
    tmp_dir.gen("m1.txt", "some content")
    scm.add(["m1.txt"])
    scm.commit("commit config file")
    tmp_dir.gen({"a": {"b": "some content"}})
    scm.add([str(tmp_dir / "a" / "b")])
    (tmp_dir / "m1.txt").unlink()
    scm.add(["m1.txt"])
    scm.commit("commit a/b")

    assert check_if_path_exists("m1.txt", scm, "HEAD^1")
    assert not check_if_path_exists("m1.txt", scm, "HEAD")
    assert not check_if_path_exists("a/b", scm, "HEAD^1")
    assert check_if_path_exists("a/b", scm, "HEAD")


def test_check_existence_no_repo(tmp_dir: TmpDir):
    tmp_dir.gen("m1.txt", "some content")
    assert check_if_path_exists(tmp_dir / "m1.txt")
    assert not check_if_path_exists(tmp_dir / "not" / "exists")



# pylint: disable=protected-access
"""TODO: add more tests for API"""
import os
from contextlib import contextmanager
from time import sleep
from typing import Optional
from unittest.mock import ANY, call, patch

import pytest
from freezegun import freeze_time
from pytest_test_utils import TmpDir
from scmrepo.git import Git

import gto
import tests.resources
from gto.api import show
from gto.exceptions import RefNotFound, WrongArgs
from gto.index import RepoIndexManager
from gto.tag import find
from gto.versions import SemVer
from tests.utils import (
    check_obj,
    convert_objects_to_str_in_json_serializable_object,
)


def test_empty_index(scm: Git):
    with RepoIndexManager.from_scm(scm) as index:
        assert isinstance(index, RepoIndexManager)
        assert len(index.artifact_centric_representation()) == 0


@pytest.mark.usefixtures("scm")
def test_empty_state(tmp_dir: TmpDir):
    state = gto.api._get_state(tmp_dir)
    assert len(state.artifacts) == 0


@pytest.mark.usefixtures("scm")
def test_api_info_commands_empty_repo(tmp_dir: TmpDir):
    gto.api.show(tmp_dir)
    gto.api.history(tmp_dir)


@pytest.fixture(name="artifact")
def _artifact(tmp_dir: TmpDir, scm: Git, request: pytest.FixtureRequest) -> str:
    request.getfixturevalue("init_showcase_semver")
    tmp_dir.gen(
        "artifacts.yaml", "rf: \n  type: model\n  path: models/random-forest.pkl\n"
    )
    scm.add(["artifacts.yaml"])
    scm.commit("Commit 1")
    tmp_dir.gen(
        "artifacts.yaml", "rf: \n  type: model\n  path: models/random-forest.pklx\n"
    )
    scm.add(["artifacts.yaml"])
    scm.commit("Commit 2")
    return "new-artifact"


# def test_api_info_commands_repo_with_artifact(
#     scm: Git,
#     artifact: str,
# ):
#     gto.api.show(scm)
#     gto.api.show(scm, artifact)
#     gto.api.history(scm)


# def test_describe(scm: Git, artifact: str):
#     gto.api.annotate(scm, artifact, path="other-path")
#     check_obj(
#         gto.api.describe(scm, artifact).dict(exclude_defaults=True),  # type: ignore
#         dict(
#             type="new-type",
#             path="other-path",
#         ),
#     )
#     check_obj(
#         gto.api.describe(scm, artifact, rev="HEAD").dict(exclude_defaults=True),  # type: ignore
#         dict(
#             type="new-type",
#             path="path",
#         ),
#     )


def test_register_deregister(tmp_dir: TmpDir, scm: Git, artifact: str):
    vname1, vname2 = "v1.0.0", "v1.0.1"
    gto.api.register(tmp_dir, artifact, "HEAD", vname1)
    latest = gto.api.find_latest_version(tmp_dir, artifact)
    assert latest.version == vname1
    tmp_dir.gen("tmp.txt", "some text")
    scm.add(["tmp.txt"])
    scm.commit("Irrelevant action to create a git commit to register another version")
    message = "Some message"
    author = "GTO"
    author_email = "gto@iterative.ai"
    gto.api.register(
        tmp_dir,
        artifact,
        "HEAD",
        message=message,
        author=author,
        author_email=author_email,
    )
    latest = gto.api.find_latest_version(tmp_dir, artifact)
    assert latest.version == vname2
    assert latest.message == message
    assert latest.author == author
    assert latest.author_email == author_email

    assert len(gto.api.show(tmp_dir, artifact, deprecated=False)) == 2

    # test _show_versions
    assert (
        gto.api._show_versions(tmp_dir, artifact, ref="HEAD")[0]["ref"]
        == f"{artifact}@v1.0.1"
    )
    assert (
        gto.api._show_versions(tmp_dir, artifact, ref="HEAD^1")[0]["ref"]
        == f"{artifact}@v1.0.0"
    )
    assert (
        gto.api._show_versions(tmp_dir, artifact, ref=scm.get_rev())[0]["ref"]
        == f"{artifact}@v1.0.1"
    )
    with pytest.raises(RefNotFound):
        gto.api._show_versions(tmp_dir, artifact, ref="HEAD^2")

    gto.api.deregister(repo=tmp_dir, name=artifact, version=vname2)
    latest = gto.api.find_latest_version(tmp_dir, artifact)
    assert latest.version == vname1

    assert len(gto.api.show(tmp_dir, artifact, deprecated=False)) == 1
    assert len(gto.api.show(tmp_dir, artifact, deprecated=True)) == 2
    assert len(gto.api._show_versions(tmp_dir, artifact, ref="HEAD")) == 0


@pytest.mark.parametrize(
    "name",
    (
        "model",
        "folder:artifact",
        "some/folder:some/artifact",
    ),
)
@pytest.mark.usefixtures("artifact")
def test_assign(tmp_dir: TmpDir, scm: Git, name: str):
    stage = "staging"
    scm.tag("v1.0.0")
    scm.tag("wrong-tag-unrelated")
    message = "some msg"
    author = "GTO"
    author_email = "gto@iterative.ai"
    event = gto.api.assign(
        tmp_dir,
        name,
        stage,
        ref="HEAD",
        name_version="v0.0.1",
        message=message,
        author=author,
        author_email=author_email,
    )
    assignments = gto.api.find_versions_in_stage(tmp_dir, name, stage)
    assert len(assignments) == 1
    check_obj(
        assignments[0].dict_state(),
        dict(
            artifact=name,
            version="v0.0.1",
            stage=stage,
            author=author,
            author_email=author_email,
            message=message,
            commit_hexsha=scm.get_rev(),
            is_active=True,
            ref=event.ref,
        ),
        {"created_at", "assignments", "unassignments", "tag", "activated_at"},
    )
    event = gto.api.assign(
        tmp_dir,
        name,
        stage,
        ref="HEAD^1",
        name_version="v0.0.2",
        message=message,
        author=author,
        author_email=author_email,
    )
    assignments = gto.api.find_versions_in_stage(tmp_dir, name, stage)
    assert len(assignments) == 1
    assignments = gto.api.find_versions_in_stage(
        tmp_dir, name, stage, versions_per_stage=-1
    )
    assert len(assignments) == 2


def test_assign_skip_registration(tmp_dir: TmpDir, artifact: str):
    stage = "staging"
    with pytest.raises(WrongArgs):
        gto.api.assign(
            tmp_dir,
            artifact,
            stage,
            ref="HEAD",
            name_version="v0.0.1",
            skip_registration=True,
        )
    gto.api.assign(tmp_dir, artifact, stage, ref="HEAD", skip_registration=True)
    assignments = gto.api.find_versions_in_stage(tmp_dir, artifact, stage)
    assert len(assignments) == 1
    assert not SemVer.is_valid(assignments[0].version)


def test_assign_force_is_needed(tmp_dir: TmpDir, artifact: str):
    gto.api.assign(tmp_dir, artifact, "staging", ref="HEAD")
    gto.api.assign(tmp_dir, artifact, "staging", ref="HEAD^1")
    with pytest.raises(WrongArgs):
        gto.api.assign(tmp_dir, artifact, "staging", ref="HEAD")
    with pytest.raises(WrongArgs):
        gto.api.assign(tmp_dir, artifact, "staging", ref="HEAD^1")
    gto.api.assign(tmp_dir, artifact, "staging", ref="HEAD", force=True)
    gto.api.assign(tmp_dir, artifact, "staging", ref="HEAD^1", force=True)


@pytest.mark.usefixtures("artifact")
def test_unassign(tmp_dir: TmpDir):
    gto.api.register(tmp_dir, name="model", ref="HEAD")
    gto.api.assign(tmp_dir, name="model", ref="HEAD", stage="dev")
    assert (
        gto.api.find_versions_in_stage(tmp_dir, name="model", stage="dev") is not None
    )

    gto.api.unassign(tmp_dir, name="model", ref="HEAD", stage="dev")
    assert gto.api.find_versions_in_stage(tmp_dir, name="model", stage="dev") is None


@pytest.mark.usefixtures("artifact")
def test_deprecate(tmp_dir: TmpDir):
    gto.api.register(tmp_dir, name="model", ref="HEAD")
    assert len(gto.api.show(tmp_dir, "model")) == 1

    sleep(1)
    gto.api.deprecate(tmp_dir, name="model")
    assert len(gto.api.show(tmp_dir, "model", deprecated=False)) == 0
    assert len(gto.api.show(tmp_dir, "model", deprecated=True)) == 1

    with pytest.raises(WrongArgs):
        gto.api.deprecate(tmp_dir, name="model")
        gto.api.deprecate(tmp_dir, name="model", simple=True, force=True)
    gto.api.deprecate(tmp_dir, name="model", force=True)


@contextmanager
def environ(**overrides):
    old = {name: os.environ[name] for name in overrides if name in os.environ}
    to_del = set(overrides) - set(old)
    try:
        os.environ.update(overrides)
        yield
    finally:
        os.environ.update(old)
        for name in to_del:
            os.environ.pop(name, None)


@pytest.mark.usefixtures("artifact")
@pytest.mark.parametrize("with_prefix", [True, False])
def test_check_ref_detailed(scm: Git, with_prefix: bool):
    NAME = "model"
    SEMVER = "v1.2.3"
    GIT_AUTHOR_NAME = "Alexander Guschin"
    GIT_AUTHOR_EMAIL = "aguschin@iterative.ai"
    GIT_COMMITTER_NAME = "Oliwav"
    GIT_COMMITTER_EMAIL = "oliwav@iterative.ai"

    with environ(
        GIT_AUTHOR_NAME=GIT_AUTHOR_NAME,
        GIT_AUTHOR_EMAIL=GIT_AUTHOR_EMAIL,
        GIT_COMMITTER_NAME=GIT_COMMITTER_NAME,
        GIT_COMMITTER_EMAIL=GIT_COMMITTER_EMAIL,
    ):
        gto.api.register(scm, name=NAME, ref="HEAD", version=SEMVER)

    ref = f"{NAME}@{SEMVER}"
    if with_prefix:
        ref = f"refs/tags/{ref}"
    events = gto.api.check_ref(scm, ref)
    assert len(events) == 1, "Should return one event"
    check_obj(
        events[0].dict_state(),
        {
            "event": "registration",
            "artifact": NAME,
            "version": SEMVER,
            "author": GIT_COMMITTER_NAME,
            "author_email": GIT_COMMITTER_EMAIL,
            "tag": f"{NAME}@{SEMVER}",
        },
        skip_keys={"commit_hexsha", "created_at", "message", "priority", "addition"},
    )


@pytest.mark.usefixtures("showcase")
def test_check_ref_multiple_showcase(scm: Git):
    for tag in find(scm=scm):
        events = gto.api.check_ref(scm, tag.name)
        assert len(events) == 1, "Should return one event"
        assert events[0].ref == tag.name


@pytest.mark.usefixtures("artifact")
def test_check_ref_catch_the_bug(scm: Git):
    NAME = "artifact"
    gto.api.register(scm, NAME, "HEAD")
    assignment1 = gto.api.assign(scm, NAME, "staging", ref="HEAD")
    assignment2 = gto.api.assign(scm, NAME, "prod", ref="HEAD")
    assignment3 = gto.api.assign(scm, NAME, "dev", ref="HEAD")
    for assignment, tag in zip(
        [assignment1, assignment2, assignment3],
        [f"{NAME}#staging#1", f"{NAME}#prod#2", f"{NAME}#dev#3"],
    ):
        events = gto.api.check_ref(scm, tag)
        assert len(events) == 1, events
        assert events[0].ref == assignment.tag == tag


@pytest.mark.usefixtures("scm")
def test_is_not_gto_repo(tmp_dir: TmpDir):
    assert not gto.api._is_gto_repo(tmp_dir)


@pytest.mark.usefixtures("init_showcase_semver")
def test_is_gto_repo_because_of_config(tmp_dir: TmpDir):
    assert gto.api._is_gto_repo(tmp_dir)


@pytest.mark.usefixtures("repo_with_commit")
def test_is_gto_repo_because_of_registered_artifact(tmp_dir: TmpDir):
    gto.api.register(tmp_dir, "model", "HEAD", "v1.0.0")
    assert gto.api._is_gto_repo(tmp_dir)


@pytest.mark.usefixtures("scm")
def test_is_gto_repo_because_of_artifacts_yaml(tmp_dir: TmpDir):
    tmp_dir.gen("artifacts.yaml", "{}")
    assert gto.api._is_gto_repo(tmp_dir)


def test_if_show_on_remote_git_repo_then_return_expected_registry():
    result = show(repo=tests.resources.SAMPLE_REMOTE_REPO_URL)
    assert result == tests.resources.get_sample_remote_repo_expected_registry()


@pytest.mark.parametrize(
    "ref,expected_stage,expected_version,expected_artifact",
    (
        ("churn#prod#2", "prod", "v3.0.0", "churn"),
        ("segment@v0.4.1", None, "v0.4.1", "segment"),
    ),
)
def test_if_check_ref_on_remote_git_repo_then_return_expected_reference(
    ref: str,
    expected_stage: Optional[str],
    expected_version: str,
    expected_artifact: str,
):
    result = gto.api.check_ref(repo=tests.resources.SAMPLE_REMOTE_REPO_URL, ref=ref)
    assert len(result) == 1
    if expected_stage is not None:
        assert result[0].stage == expected_stage
    else:
        assert hasattr(result[0], "stage") is False
    assert result[0].version == expected_version
    assert result[0].artifact == expected_artifact


@freeze_time("1996-06-09 00:00:00", tz_offset=0)
def test_if_history_on_remote_git_repo_then_return_expected_history():
    result = gto.api.history(
        repo=tests.resources.SAMPLE_REMOTE_REPO_URL, artifact="churn"
    )
    assert (
        convert_objects_to_str_in_json_serializable_object(result)
        == tests.resources.get_sample_remote_repo_expected_history_churn()
    )


def test_if_stages_on_remote_git_repo_then_return_expected_stages():
    result = gto.api.get_stages(repo=tests.resources.SAMPLE_REMOTE_REPO_URL)
    assert result == ["dev", "prod", "staging"]


def test_if_describe_on_remote_git_repo_then_return_expected_info():
    result = gto.api.describe(repo=tests.resources.SAMPLE_REMOTE_REPO_URL, name="churn")
    assert result.dict(exclude_defaults=True) == {
        "type": "model",
        "path": "models/churn.pkl",
        "virtual": False,
    }


@pytest.mark.usefixtures("artifact")
def test_if_register_with_auto_push_then_invoke_git_push_tag(tmp_dir: TmpDir):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.register(repo=tmp_dir, name="model", ref="HEAD", push=True)
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY,
        tag_name="model@v0.0.1",
        delete=False,
    )


@pytest.mark.usefixtures("artifact")
def test_if_assign_with_auto_push_then_invoke_git_push_tag_2_times_for_registration_and_promotion(
    tmp_dir: TmpDir,
):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.assign(tmp_dir, name="model", stage="dev", ref="HEAD", push=True)
    expected_calls = [
        call(
            scm=ANY,
            tag_name="model@v0.0.1",
            delete=False,
        ),
        call(
            scm=ANY,
            tag_name="model#dev#1",
            delete=False,
        ),
    ]
    mocked_git_push_tags.assert_has_calls(expected_calls)


@pytest.mark.usefixtures("artifact")
def test_if_unassign_with_auto_push_then_invoke_git_push_tag(tmp_dir: TmpDir):
    gto.api.assign(tmp_dir, name="model", stage="dev", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.unassign(
            tmp_dir,
            name="model",
            stage="dev",
            version="v0.0.1",
            push=True,
        )
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY,
        tag_name="model#dev!#2",
        delete=False,
    )


@pytest.mark.usefixtures("artifact")
def test_if_unassign_with_delete_and_auto_push_then_invoke_git_push_tag(
    tmp_dir: TmpDir,
):
    gto.api.assign(tmp_dir, name="model", stage="dev", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.unassign(
            tmp_dir,
            name="model",
            stage="dev",
            version="v0.0.1",
            delete=True,
            push=True,
        )
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY, tag_name="model#dev#1", delete=True
    )


@pytest.mark.usefixtures("artifact")
def test_if_deregister_with_auto_push_then_invoke_git_push_tag(tmp_dir: TmpDir):
    gto.api.register(tmp_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deregister(tmp_dir, name="model", version="v0.0.1", push=True)
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY,
        tag_name="model@v0.0.1!",
        delete=False,
    )


@pytest.mark.usefixtures("artifact")
def test_if_deregister_with_delete_and_auto_push_then_invoke_git_push_tag(
    tmp_dir: TmpDir,
):
    gto.api.register(tmp_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deregister(
            tmp_dir,
            name="model",
            version="v0.0.1",
            push=True,
            delete=True,
        )
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY,
        tag_name="model@v0.0.1",
        delete=True,
    )


@pytest.mark.usefixtures("artifact")
def test_if_deprecate_with_auto_push_then_invoke_git_push_tag(tmp_dir: TmpDir):
    gto.api.register(tmp_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deprecate(tmp_dir, name="model", push=True)
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY,
        tag_name="model@deprecated",
        delete=False,
    )


@pytest.mark.usefixtures("artifact")
def test_if_deprecate_with_delete_and_auto_push_then_invoke_git_push_tag(
    tmp_dir: TmpDir,
):
    gto.api.register(tmp_dir, name="model", ref="HEAD", push=False)
    with patch("gto.registry.git_push_tag") as mocked_git_push_tags:
        gto.api.deprecate(tmp_dir, name="model", push=True, delete=True)
    mocked_git_push_tags.assert_called_once_with(
        scm=ANY,
        tag_name="model@v0.0.1",
        delete=True,
    )


def test_if_register_with_remote_repo_then_invoke_git_push_tag(tmp_dir: TmpDir):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.register(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="model",
                ref="HEAD",
            )
            mocked_git_push_tag.assert_called_once_with(
                scm=ANY,
                tag_name="model@v0.0.1",
                delete=False,
            )


def test_if_assign_with_remote_repo_then_invoke_git_push_tag(tmp_dir: TmpDir):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.assign(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="model",
                stage="dev",
                ref="HEAD",
            )
            expected_calls = [
                call(
                    scm=ANY,
                    tag_name="model@v0.0.1",
                    delete=False,
                ),
                call(
                    scm=ANY,
                    tag_name="model#dev#1",
                    delete=False,
                ),
            ]
            mocked_git_push_tag.assert_has_calls(expected_calls)


def test_if_deprecate_with_remote_repo_then_invoke_git_push_tag(tmp_dir: TmpDir):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.deprecate(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="churn",
            )
            mocked_git_push_tag.assert_called_once_with(
                scm=ANY,
                tag_name="churn@deprecated",
                delete=False,
            )


def test_if_deregister_with_remote_repo_then_invoke_git_push_tag(tmp_dir: TmpDir):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.deregister(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="churn",
                version="v3.0.0",
            )
            mocked_git_push_tag.assert_called_once_with(
                scm=ANY,
                tag_name="churn@v3.0.0!",
                delete=False,
            )


def test_if_unassign_with_remote_repo_then_invoke_git_push_tag(tmp_dir: TmpDir):
    with patch("gto.registry.git_push_tag") as mocked_git_push_tag:
        with patch("gto.git_utils.TemporaryDirectory") as MockedTemporaryDirectory:
            MockedTemporaryDirectory.return_value = tmp_dir
            gto.api.unassign(
                repo=tests.resources.SAMPLE_REMOTE_REPO_URL,
                name="churn",
                stage="staging",
                version="v3.1.0",
            )
            mocked_git_push_tag.assert_called_once_with(
                scm=ANY,
                tag_name="churn#staging!#3",
                delete=False,
            )



import pytest
from scmrepo.git import Git

from gto.constants import Action
from gto.exceptions import RefNotFound, TagExists
from gto.tag import create_tag, find, name_tag, parse_name


def test_name_tag(scm: Git):
    assert name_tag(Action.REGISTER, "myartifact", "v1", simple=True) == "myartifact@v1"
    assert (
        name_tag(Action.REGISTER, "myartifact", "v1", simple=False, scm=scm)
        == "myartifact@v1#1"
    )
    assert (
        name_tag(Action.DEREGISTER, "myartifact", "v1", simple=True) == "myartifact@v1!"
    )
    assert (
        name_tag(Action.DEREGISTER, "myartifact", "v1", simple=False, scm=scm)
        == "myartifact@v1!#1"
    )
    assert (
        name_tag(Action.ASSIGN, "myartifact", stage="stage", simple=True)
        == "myartifact#stage"
    )
    assert (
        name_tag(Action.ASSIGN, "myartifact", scm=scm, stage="stage", simple=False)
        == "myartifact#stage#1"
    )
    assert (
        name_tag(Action.UNASSIGN, "myartifact", scm=scm, stage="stage", simple=False)
        == "myartifact#stage!#1"
    )


def test_parse_name():
    assert parse_name("path@v1.2.3") == dict(
        name="path", version="v1.2.3", action=Action.REGISTER
    )
    assert parse_name("path@v1.2.3#5") == dict(
        name="path", version="v1.2.3", action=Action.REGISTER, counter=5
    )
    assert parse_name("path@v1.2.3!") == dict(
        name="path", version="v1.2.3", action=Action.DEREGISTER
    )
    assert parse_name("path@v1.2.3!#2") == dict(
        name="path", version="v1.2.3", action=Action.DEREGISTER, counter=2
    )
    assert parse_name("path#stage") == dict(
        name="path", action=Action.ASSIGN, stage="stage"
    )
    assert parse_name("path#stage#1") == dict(
        name="path", action=Action.ASSIGN, stage="stage", counter=1
    )
    assert parse_name("path#stage!#2") == dict(
        name="path", action=Action.UNASSIGN, stage="stage", counter=2
    )


@pytest.mark.parametrize(
    "tag_name",
    [
        "",
        "###",
        "@@@",
        "model@v1",
        "nn#prod#-1",
        "model@v111",
        "model-prod",
        "model@0.0.1",
        "model-prod-v1",
        "model#prod#-1",
        "model#prod#1-2",
        "model#prod?#1-2",
        "model#prod@v1.2.3",
        "model@v1.2.3#prod",
        "model-prod-v1-stage",
        "model@prod@v1@stage-1",
        "namespace/model@0.0.1",
        "@namespace/model@0.0.1",
        "model#prod#v1#stage#1-",
        "model#prod-v1-stage@1-2",
        "model#prod-v1-stage@1-2#@",
    ],
)
def test_parse_wrong_names(tag_name: str):
    assert not parse_name(tag_name, raise_on_fail=False)


@pytest.mark.parametrize(
    "tag_name",
    [
        "nn#prod",
        "nn#prod!",
        "nn#prod#1",
        "nn@v1.0.0",
        "nn#prod!#2",
        "nn@v1.0.0!",
        "nn@v1.0.0#1",
        "nn@v1.0.0!#1",
        "nn@v1.0.0-rc.2",
        "nn@v1.0.0-alpha",
        "nn@v1.0.0-alpha.1",
        "nn@v1.0.0-alpha.beta",
        "new-artifact@v1.0.1!",
    ],
)
def test_parse_correct_names(tag_name: str):
    assert parse_name(tag_name, raise_on_fail=False)


@pytest.mark.usefixtures("repo_with_commit")
def test_create_tag_bad_ref(scm: Git):
    with pytest.raises(RefNotFound):
        create_tag(scm, "name", rev="wrongref", message="msg")
    with pytest.raises(RefNotFound):
        create_tag(
            scm, "name", rev="679dd96f8f22bef6505b9646803bf3c2afe94692", message="msg"
        )


@pytest.mark.usefixtures("repo_with_commit")
def test_create_tag_repeated_tagname(scm: Git):
    create_tag(scm, "name", rev="HEAD", message="msg")
    with pytest.raises(TagExists):
        create_tag(scm, "name", rev="HEAD", message="msg")


@pytest.mark.usefixtures("repo_with_commit")
def test_lightweight_tag(scm: Git):
    scm.tag("lightweight-tag@v0.0.1")
    assert find(scm=scm) == []



import logging
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import (
    IO,
    Any,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Union,
)

from pydantic import BaseModel, ValidationError, parse_obj_as, validator
from ruamel.yaml import YAMLError
from scmrepo.exceptions import SCMError
from scmrepo.git import Git
from scmrepo.git.objects import GitCommit

from gto.base import BaseManager, BaseRegistryState
from gto.base import Commit as EnrichmentEvent
from gto.commit_message_generator import (
    generate_annotate_commit_message,
    generate_remove_commit_message,
)
from gto.config import (
    CONFIG_FILE_NAME,
    RegistryConfig,
    assert_name_is_valid,
    read_registry_config,
    yaml,
)
from gto.constants import Action
from gto.exceptions import (
    ArtifactExists,
    ArtifactNotFound,
    NoFile,
    PathIsUsed,
    WrongArgs,
    WrongArtifactsYaml,
)
from gto.ext import EnrichmentInfo, EnrichmentReader
from gto.git_utils import RemoteRepoMixin
from gto.ui import echo

logger = logging.getLogger("gto")


class Artifact(BaseModel):
    type: Optional[str] = None
    path: Optional[str] = None
    virtual: bool = True
    labels: List[str] = []  # TODO: allow key:value labels
    description: str = ""
    custom: Any = None


State = Dict[str, Artifact]


def not_frozen(func):
    @wraps(func)
    def inner(self: "Index", *args, **kwargs):
        if self.frozen:
            raise ValueError(f"Cannot {func.__name__}: {self.__class__} is frozen")
        return func(self, *args, **kwargs)

    return inner


def find_repeated_path(
    path: Union[str, Path], paths: Iterable[Union[str, Path]]
) -> Optional[Path]:
    """Return the path from "paths" that conflicts with "path":
    is equal to it or is a subpath (in both directions).
    """
    path = Path(path).resolve()
    for p in paths:
        if p:
            p = Path(p).resolve()
            if p == path or p in path.parents or path in p.parents:
                return p
    return None


def check_if_path_exists(
    path: Union[str, Path], scm: Optional[Git] = None, ref: str = None
):
    """Check if path was committed to repo
    or it just exists in case the repo is not provided.
    """
    if scm is None:
        return Path(path).exists()
    try:
        fs = scm.get_fs(ref)
        return fs.exists(str(path))
    except SCMError:
        return False


class Index(BaseModel):
    state: State = {}  # TODO should not be populated until load() is called
    frozen: bool = False

    @validator("state")
    def state_is_valid(cls, v):  # pylint: disable=no-self-argument, no-self-use
        for name, artifact in v.items():
            assert_name_is_valid(name)
            if artifact.type:
                assert_name_is_valid(artifact.type)
            for label in artifact.labels:
                assert_name_is_valid(label)
        return v

    def __contains__(self, item):
        return item in self.state

    @classmethod
    def read(cls, path_or_file: Union[str, IO], frozen: bool = False):
        index = cls(frozen=frozen)
        index.state = index.read_state(path_or_file)
        return index

    @staticmethod
    def read_state(path_or_file: Union[str, IO]):
        def read_yaml(stream: IO):
            try:
                return yaml.load(stream)
            except YAMLError as e:
                raise WrongArtifactsYaml() from e

        # read contents of the yaml
        if isinstance(path_or_file, str):
            try:
                with open(path_or_file, "r", encoding="utf8") as file:
                    contents = read_yaml(file)
            except FileNotFoundError as e:
                raise NoFile("artifacts.yaml") from e
        else:
            contents = read_yaml(path_or_file)
        # check yaml contents is a valid State
        try:
            state = parse_obj_as(State, contents)
        except ValidationError as e:
            raise WrongArtifactsYaml() from e
        # validate that specific names conform to the naming convention
        for key in state:
            assert_name_is_valid(key)
        return state

    def write_state(self, path_or_file: Union[str, IO]):
        if isinstance(path_or_file, str):
            with open(path_or_file, "w", encoding="utf8") as file:
                state = self.dict(exclude_defaults=True).get("state", {})
                yaml.dump(state, file)

    @not_frozen
    def add(
        self,
        name,
        type,
        path,
        must_exist,
        allow_same_path,
        labels,
        description,
        custom,
        update,
    ) -> Artifact:
        if name in self and not update:
            raise ArtifactExists(name)
        if (
            path
            and not allow_same_path
            and find_repeated_path(
                path, [a.path for n, a in self.state.items() if n != name]
            )
            is not None
        ):
            raise PathIsUsed(type=type, name=name, path=path)
        if update and name in self.state:
            self.state[name].type = type or self.state[name].type
            self.state[name].path = path or self.state[name].path
            if must_exist:
                self.state[name].virtual = False
            elif path:
                self.state[name].virtual = True
            self.state[name].labels = sorted(set(self.state[name].labels).union(labels))
            self.state[name].description = description or self.state[name].description
            self.state[name].custom = custom or self.state[name].custom
        else:
            self.state[name] = Artifact(
                type=type,
                path=path,
                virtual=not must_exist,
                labels=labels,
                description=description,
                custom=custom,
            )
        self.state_is_valid(self.state)  # type: ignore[call-arg]
        return self.state[name]

    @not_frozen
    def remove(self, name):
        if name not in self:
            raise ArtifactNotFound(name)
        del self.state[name]


class BaseIndexManager(BaseModel, ABC):
    current: Optional[Index]
    config: RegistryConfig

    @abstractmethod
    def get_index(self) -> Index:
        raise NotImplementedError

    @abstractmethod
    def update(self):
        raise NotImplementedError

    @abstractmethod
    def get_history(self) -> Dict[str, Index]:
        raise NotImplementedError

    def add(
        self,
        name,
        type,
        path,
        must_exist,
        allow_same_path,
        labels,
        description,
        custom,
        update,
        stdout=False,
    ):
        for arg in [name] + list(labels or []):
            assert_name_is_valid(arg)
        if type:
            self.config.assert_type(type)
        if must_exist:
            if not path:
                raise WrongArgs("`path` is required when `must_exist` is set to True")
            if not check_if_path_exists(
                path, self.scm if hasattr(self, "scm") else None
            ):
                raise NoFile(path)
        index = self.get_index()
        index.add(
            name,
            type=type,
            path=path,
            must_exist=must_exist,
            allow_same_path=allow_same_path,
            labels=labels or [],
            description=description,
            custom=custom,
            update=update,
        )
        self.update()
        if stdout:
            echo("Updated `artifacts.yaml`")

    def remove(self, name, stdout=False):
        index = self.get_index()
        index.remove(name)
        self.update()
        if stdout:
            echo("Updated `artifacts.yaml`")


class FileIndexManager(BaseIndexManager):
    path: str = ""

    @classmethod
    def from_path(cls, path: str, config: RegistryConfig = None):
        if config is None:
            config = read_registry_config(os.path.join(path, CONFIG_FILE_NAME))
        return cls(path=path, config=config)

    def index_path(self):
        return str(Path(self.path) / self.config.INDEX)

    def get_index(self) -> Index:
        if os.path.exists(self.index_path()):
            self.current = Index.read(self.index_path())
        if not self.current:
            self.current = Index()
        return self.current

    def update(self):
        if self.current is not None:
            self.current.write_state(self.index_path())

    def get_history(self) -> Dict[str, Index]:
        raise NotImplementedError("Not a git repo: history is not available")

    def get_commit_index(self, **kwargs) -> Index:
        raise NotImplementedError("Not a git repo: using revision is not available")

    def artifact_centric_representation(self):
        raise NotImplementedError("Not a git repo")


ArtifactCommits = Dict[str, Artifact]
ArtifactsCommits = Dict[str, ArtifactCommits]


class RepoIndexManager(FileIndexManager, RemoteRepoMixin):
    scm: Git

    def __init__(self, scm: Git, config):
        super().__init__(scm=scm, config=config)

    @classmethod
    @contextmanager
    def from_scm(cls, scm: Git, config: RegistryConfig = None):
        if config is None:
            config = read_registry_config(os.path.join(scm.root_dir, CONFIG_FILE_NAME))
        yield cls(scm=scm, config=config)

    def add(
        self,
        name,
        type,
        path,
        must_exist,
        allow_same_path,
        labels,
        description,
        custom,
        update,
        stdout=False,
        commit=False,
        commit_message=None,
        push=False,
    ):
        return self._call_commit_push(
            super().add,
            commit=commit,
            commit_message=commit_message
            or generate_annotate_commit_message(name=name, type=type, path=path),
            push=push,
            stdout=stdout,
            name=name,
            type=type,
            path=path,
            must_exist=must_exist,
            allow_same_path=allow_same_path,
            labels=labels,
            description=description,
            custom=custom,
            update=update,
        )

    def remove(
        self,
        name,
        stdout=False,
        commit=False,
        commit_message=None,
        push=False,
    ):
        return self._call_commit_push(
            super().remove,
            commit=commit,
            commit_message=commit_message or generate_remove_commit_message(name=name),
            push=push,
            stdout=stdout,
            name=name,
        )

    def index_path(self):
        # TODO: config should be loaded from repo too
        return os.path.join(self.scm.root_dir, self.config.INDEX)

    class Config:
        arbitrary_types_allowed = True

    def get_commit_index(  # type: ignore # pylint: disable=arguments-differ
        self,
        rev: Union[str, GitCommit],
        allow_to_not_exist: bool = True,
        ignore_corrupted: bool = False,
    ) -> Optional[Index]:
        rev = rev.hexsha if isinstance(rev, GitCommit) else self.scm.resolve_rev(rev)
        return self._get_commit_index(
            rev,
            allow_to_not_exist=allow_to_not_exist,
            ignore_corrupted=ignore_corrupted,
        )

    def _get_commit_index(  # type: ignore # pylint: disable=arguments-differ
        self,
        rev: str,
        allow_to_not_exist: bool = True,
        ignore_corrupted: bool = False,
    ) -> Optional[Index]:
        fs = self.scm.get_fs(rev)
        try:
            with fs.open(self.config.INDEX) as f:
                try:
                    return Index.read(f, frozen=True)
                except WrongArtifactsYaml as e:
                    logger.warning("Corrupted artifacts.yaml file in commit %s", rev)
                    if ignore_corrupted:
                        return None
                    raise e
        except FileNotFoundError:
            if allow_to_not_exist:
                return None
        raise ValueError(f"No Index exists at {rev}")

    def get_history(self) -> Dict[str, Index]:
        revs = {
            rev
            for branch in self.scm.iter_refs("refs/heads/")
            for rev in self.scm.branch_revs(branch)
        }
        history = {}
        for rev in revs:
            try:
                index = self._get_commit_index(rev)
                if index is not None:
                    history[rev] = index
            except ValueError:
                pass
        return history

    def artifact_centric_representation(self) -> ArtifactsCommits:
        representation = defaultdict(dict)  # type: ArtifactsCommits
        for commit, index in self.get_history().items():
            for art in index.state:
                representation[art][commit] = index.state[art]
        return representation

    def check_existence(self, name, commit):
        return name in self.get_commit_index(commit)

    def assert_existence(self, name, commit):
        if not self.check_existence(name, commit):
            raise ArtifactNotFound(name)


class EnrichmentManager(BaseManager, RemoteRepoMixin):
    actions: FrozenSet[Action] = frozenset()

    @classmethod
    @contextmanager
    def from_scm(cls, scm: Git, config: RegistryConfig = None):
        if config is None:
            config = read_registry_config(os.path.join(scm.root_dir, CONFIG_FILE_NAME))
        yield cls(scm=scm, config=config)

    def describe(self, name: str, rev: str = None) -> List[EnrichmentInfo]:
        enrichments = self.config.enrichments
        res = []
        gto_enrichment = enrichments.pop("gto")
        gto_info = gto_enrichment.describe(self.scm, name, rev)
        if gto_info:
            res.append(gto_info)
            path = gto_info.get_path()  # type: ignore
            for enrichment in enrichments.values():
                enrichment_data = enrichment.describe(self.scm, path, rev)
                if enrichment_data is not None:
                    res.append(enrichment_data)
        return res

    def get_commits(self, all_branches=False, all_commits=False) -> Iterator[GitCommit]:
        revs: Set[str] = set()
        if all_commits or all_branches:
            branches = list(self.scm.iter_refs("refs/heads/"))
            if all_commits:
                revs.update(
                    rev for branch in branches for rev in self.scm.branch_revs(branch)
                )
            elif all_branches:
                revs.update(self.scm.resolve_rev(branch) for branch in branches)
        else:
            try:
                revs.add(self.scm.get_rev())
            except SCMError:
                pass  # empty git repo
        yield from (self.scm.resolve_commit(rev) for rev in revs)

    def update_state(
        self,
        state: BaseRegistryState,
        all_branches=False,
        all_commits=False,
    ) -> BaseRegistryState:
        # processing registered artifacts and versions first
        for artifact in state.get_artifacts().values():
            for version in artifact.versions:
                commit = self.scm.resolve_commit(version.commit_hexsha)
                enrichments = self.describe(
                    artifact.artifact,
                    # faster to make git.Reference here
                    rev=commit,
                )
                version.add_event(
                    EnrichmentEvent(
                        artifact=artifact.artifact,
                        version=version.version,
                        created_at=datetime.fromtimestamp(commit.commit_time),
                        author=commit.author_name,
                        author_email=commit.author_email,
                        commit_hexsha=commit.hexsha,
                        message=commit.message,
                        committer=commit.committer_name,
                        committer_email=commit.committer_email,
                        enrichments=enrichments,
                    )
                )
                state.update_artifact(artifact)
        for commit in self.get_commits(
            all_branches=all_branches, all_commits=all_commits
        ):
            for art_name in GTOEnrichment().discover(self.scm, commit):
                enrichments = self.describe(art_name, rev=commit)
                artifact = state.find_artifact(art_name, create_new=True)
                version = artifact.find_version(
                    commit_hexsha=commit.hexsha, create_new=True
                )
                version.add_event(
                    EnrichmentEvent(
                        artifact=artifact.artifact,
                        version=version.version,
                        created_at=datetime.fromtimestamp(commit.commit_time),
                        author=commit.author_name,
                        author_email=commit.author_email,
                        commit_hexsha=commit.hexsha,
                        message=commit.message,
                        committer=commit.committer_name,
                        committer_email=commit.committer_email,
                        enrichments=enrichments,
                    )
                )
                state.update_artifact(artifact)
        return state


class GTOInfo(EnrichmentInfo):
    source = "gto"
    artifact: Artifact

    def get_object(self) -> BaseModel:
        return self.artifact

    def get_human_readable(self) -> str:
        return self.artifact.json()

    def get_path(self):
        return self.artifact.path


class GTOEnrichment(EnrichmentReader):
    source = "gto"

    def discover(  # pylint: disable=no-self-use
        self, url_or_scm: Union[str, Git], rev: str
    ) -> Dict[str, GTOInfo]:
        with RepoIndexManager.from_url(url_or_scm) as index:
            index = index.get_commit_index(rev)
        if index:
            return {
                name: GTOInfo(artifact=artifact)
                for name, artifact in index.state.items()
            }
        return {}

    def describe(  # pylint: disable=no-self-use
        self, url_or_scm: Union[str, Git], obj: str, rev: Optional[str]
    ) -> Optional[GTOInfo]:
        with RepoIndexManager.from_url(url_or_scm) as index:
            index = index.get_commit_index(rev or index.scm.get_rev())
        if index and obj in index.state:
            return GTOInfo(artifact=index.state[obj])
        return None



import gto.log  # noqa
from gto import api
from gto._version import __version__
from gto.config import CONFIG
from gto.index import RepoIndexManager
from gto.registry import GitRegistry

__all__ = ["api", "CONFIG", "RepoIndexManager", "GitRegistry", "__version__"]



# pylint: disable=unused-variable, redefined-outer-name
from typing import Callable, Optional, Tuple
from unittest import mock

import pytest
import typer
from packaging import version
from pytest_test_utils import TmpDir
from typer.main import get_command_from_info

from gto.cli import app
from gto.exceptions import GTOException
from tests.conftest import Runner


def _check_output_contains(output: str, search_value: str) -> bool:
    return search_value in output


def _check_output_exact_match(output: str, search_value: str) -> bool:
    return search_value == output


def _check_successful_cmd(
    cmd: str,
    args: list,
    expected_stdout: Optional[str],
    search_func: Optional[Callable] = _check_output_exact_match,
):
    runner = Runner()
    result = runner.invoke([cmd] + args)
    assert result.exit_code == 0, (result.stdout, result.stderr, result.exception)
    if expected_stdout is not None:
        if len(expected_stdout):
            assert len(result.stdout) > 0, "Output is empty, but should not be"
        assert search_func(result.stdout, expected_stdout)


def _check_failing_cmd(
    cmd: str,
    args: list,
    expected_stderr: str,
    search_func: Optional[Callable] = _check_output_exact_match,
):
    runner = Runner()
    result = runner.invoke([cmd] + args)
    assert result.exit_code != 0, (result.stdout, result.stderr, result.exception)
    if expected_stderr is not None:
        assert len(result.stderr) > 0, "Output is empty, but should not be"
    assert search_func(result.stderr, expected_stderr)


@pytest.fixture
def app_cmd():
    return app.registered_commands


@pytest.fixture
def app_cli_cmd(app_cmd):
    if version.parse(typer.__version__) < version.parse("0.6.0"):
        return (
            get_command_from_info(c) for c in app_cmd  # pylint: disable=missing-kwoa
        )
    return (
        get_command_from_info(  # pylint: disable=unexpected-keyword-arg
            c,
            pretty_exceptions_short=False,
            rich_markup_mode="rich",
        )
        for c in app_cmd
    )


def test_commands_help(app_cli_cmd):
    no_help = [cli_cmd.name for cli_cmd in app_cli_cmd if cli_cmd.help is None]
    assert not no_help, f"{no_help} cli command do not have help!"


def test_commands_args_help(app_cli_cmd):
    no_help = []
    for cmd in app_cli_cmd:
        no_help.extend(
            f"{cmd.name}:{arg.name}"
            for arg in cmd.params
            if arg.help is None or arg.help == ""
        )

    assert not no_help, f"{no_help} cli command args do not have help!"


def test_show(empty_git_repo: str):
    _check_successful_cmd(
        "show",
        ["-r", empty_git_repo],
        "Nothing found in the current workspace\n",
    )
    _check_successful_cmd(
        "history",
        ["-r", empty_git_repo],
        "Nothing found in the current workspace\n",
    )


EXPECTED_DESCRIBE_OUTPUT = """{
    "type": "model",
    "path": "models/random-forest.pkl",
    "virtual": false
}
"""


# this is one function because showcase fixture takes some time to be created
def test_commands(tmp_dir: TmpDir, showcase: Tuple[str, str]):
    first_commit, second_commit = showcase
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir, "rf@greatest", "--version"],
        "v1.2.4\n",
    )
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir, "rf@latest", "--ref"],
        "rf@v1.2.4\n",
    )
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir, "rf#production", "--version"],
        "v1.2.3\n",
    )
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir, "rf#production", "--vs", "-1", "--version"],
        "v1.2.4\nv1.2.3\n",
    )
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir, "rf#staging", "--vs", "-1", "--version"],
        "v1.2.4\n",
    )
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir, "rf#production", "--ref"],
        "rf@v1.2.3\n",
    )
    # None because of random order - fix this
    _check_successful_cmd("stages", ["-r", tmp_dir], None)
    # None because of output randomness and complexity
    _check_successful_cmd(
        "show",
        ["-r", tmp_dir],
        None,
    )
    # None because of output randomness and complexity
    _check_successful_cmd(
        "history",
        ["-r", tmp_dir],
        None,
    )
    # check-ref
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf#production#3", "--name"],
        "rf\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "refs/tags/rf#production#3", "--name"],
        "rf\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf#production#3", "--stage"],
        "production\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf#production#3", "--version"],
        "v1.2.4\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf@v1.2.4", "--version"],
        "v1.2.4\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf@v1.2.4", "--event"],
        "registration\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf@v1.2.4"],
        '✅  Version "v1.2.4" of artifact "rf" was registered\n',
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf#production#3", "--event"],
        "assignment\n",
    )
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "rf#production#3"],
        '✅  Stage "production" was assigned to version "v1.2.4" of artifact "rf"\n',
    )
    # TODO: make unsuccessful
    _check_successful_cmd(
        "check-ref",
        ["-r", tmp_dir, "this-tag-does-not-exist"],
        "",
    )
    _check_successful_cmd(
        "parse-tag",
        ["dvclive/dsd=nn@v0.0.1"],
        "dvclive/dsd:nn",
        _check_output_contains,
    )
    _check_successful_cmd(
        "doctor",
        ["-r", tmp_dir],
        None,
    )


EXPECTED_DESCRIBE_OUTPUT_2 = """{
    "type": "new-type",
    "path": "new/path",
    "labels": [
        "another-label",
        "new-label",
        "some-label"
    ],
    "description": "new description"
}
"""


def test_register(repo_with_commit: str):
    _check_successful_cmd(
        "register",
        ["-r", repo_with_commit, "a1"],
        "Created git tag 'a1@v0.0.1' that registers version\n"
        "To push the changes upstream, run:\n"
        "    git push origin a1@v0.0.1\n",
    )

    _check_successful_cmd(
        "deprecate",
        ["-r", repo_with_commit, "a1", "v0.0.1", "--delete"],
        "Deleted git tag 'a1@v0.0.1'\n"
        "To push the changes upstream, run:\n"
        "    git push --delete origin a1@v0.0.1\n",
    )

    _check_successful_cmd(
        "register",
        ["-r", repo_with_commit, "a2", "--version", "v1.2.3"],
        "Created git tag 'a2@v1.2.3' that registers version\n"
        "To push the changes upstream, run:\n"
        "    git push origin a2@v1.2.3\n",
    )

    _check_successful_cmd(
        "deprecate",
        ["-r", repo_with_commit, "a2", "v1.2.3"],
        "Created git tag 'a2@v1.2.3!' that deregisters version\n"
        "To push the changes upstream, run:\n"
        "    git push origin a2@v1.2.3!\n",
    )

    _check_successful_cmd(
        "register",
        ["-r", repo_with_commit, "a2", "--simple", "false"],
        "Created git tag 'a2@v1.2.3#1' that registers version\n"
        "To push the changes upstream, run:\n"
        "    git push origin a2@v1.2.3#1\n",
    )

    _check_failing_cmd(
        "register",
        ["-r", repo_with_commit, "a3", "--version", "1.2.3"],
        "❌ Supplied version '1.2.3' cannot be parsed\n",
    )

    _check_successful_cmd(
        "register",
        ["-r", repo_with_commit, "classification/dvclive:models/nn"],
        "classification/dvclive=models/nn@v0.0.1",
        search_func=_check_output_contains,
    )


def test_assign(repo_with_commit: str):
    _check_successful_cmd(
        "register",
        ["-r", repo_with_commit, "nn1"],
        "Created git tag 'nn1@v0.0.1' that registers version\n"
        "To push the changes upstream, run:\n"
        "    git push origin nn1@v0.0.1\n",
    )
    # this check depends on the previous one
    _check_successful_cmd(
        "assign",
        ["-r", repo_with_commit, "nn1", "HEAD", "--stage", "prod"],
        "Created git tag 'nn1#prod#1' that assigns stage to version 'v0.0.1'\n"
        "To push the changes upstream, run:\n"
        "    git push origin nn1#prod#1\n",
    )
    # this check depends on the previous assignment
    _check_failing_cmd(
        "assign",
        [
            "-r",
            repo_with_commit,
            "nn1",
            "HEAD",
            "--version",
            "v1.0.0",
            "--stage",
            "stage",
        ],
        "❌ Can't register 'v1.0.0', since 'v0.0.1' is registered already at this ref\n",
    )

    _check_failing_cmd(
        "assign",
        [
            "-r",
            repo_with_commit,
            "nn2",
            "HEAD",
            "--version",
            "1.0.0",
            "--stage",
            "prod",
        ],
        "❌ Version '1.0.0' is not valid. Example of valid version: 'v1.0.0'\n",
    )

    _check_successful_cmd(
        "assign",
        ["-r", repo_with_commit, "nn2", "HEAD", "--stage", "prod"],
        "Created git tag 'nn2@v0.0.1' that registers version\n"
        "To push the changes upstream, run:\n"
        "    git push origin nn2@v0.0.1\n"
        "Created git tag 'nn2#prod#1' that assigns stage to version 'v0.0.1'\n"
        "To push the changes upstream, run:\n"
        "    git push origin nn2#prod#1\n",
    )


GTO_EXCEPTION_MESSAGE = "Test GTOException Message"


def test_stderr_gto_exception():
    # patch gto show to throw gto exception.
    with mock.patch("gto.api.show", side_effect=GTOException(GTO_EXCEPTION_MESSAGE)):
        _check_failing_cmd("show", [], GTO_EXCEPTION_MESSAGE, _check_output_contains)


EXCEPTION_MESSAGE = "Test Exception Message"


def test_stderr_exception():
    # patch gto show to throw exception.
    with mock.patch("gto.api.show", side_effect=Exception(EXCEPTION_MESSAGE)):
        _check_failing_cmd("show", [], EXCEPTION_MESSAGE, _check_output_contains)



"""TODO: break this file into multiple test/files"""
# pylint: disable=too-many-locals, too-many-statements
from typing import Tuple

from pytest_test_utils import TmpDir
from scmrepo.git import Git

import gto
from gto.base import (
    Artifact,
    Assignment,
    Commit,
    Deregistration,
    Registration,
    Unassignment,
    Version,
    VStage,
)
from tests.utils import check_obj


def test_api(tmp_dir: TmpDir, scm: Git, showcase: Tuple[str, str]):
    first_commit, second_commit = showcase

    gto.api.show(scm)
    gto.api.history(scm)
    for name in "nn", "rf", "features":
        gto.api.show(scm, name)
        gto.api.history(scm, name)

    artifacts = gto.api._get_state(  # pylint: disable=protected-access
        tmp_dir
    ).artifacts
    assert set(artifacts.keys()) == {"nn", "rf", "features"}
    # assert isinstance(artifacts["features"], BaseArtifact)
    # _check_obj(
    #     artifacts["features"],
    #     dict(name="features", versions=[]),
    #     ["commits"],
    # )
    nn_artifact = artifacts["nn"]
    assert isinstance(nn_artifact, Artifact)
    assert nn_artifact.artifact == "nn"
    assert len(nn_artifact.versions) == 2
    nn_version = nn_artifact.versions[0]
    assert isinstance(nn_version, Version)
    commit = scm.resolve_commit("HEAD")
    author = commit.author_name
    author_email = commit.author_email

    skip_keys_registration = {
        "created_at",
        "activated_at",
        "registrations",
        "deregistrations",
        "enrichments",
        "tag",
        "message",
        "stages",
    }
    skip_keys_assignment = {
        "created_at",
        "message",
        "assignments",
        "unassignments",
        "activated_at",
        "priority",
        "addition",
        "event",
    }

    check_obj(
        nn_version.dict_state(),
        dict(
            artifact="nn",
            version="v0.0.1",
            author=author,
            author_email=author_email,
            commit_hexsha=first_commit,
            discovered=False,
            is_active=True,
            ref="nn@v0.0.1",
        ),
        skip_keys=skip_keys_registration,
    )
    nn_vstages = nn_artifact.get_vstages()
    assert len(nn_vstages) == 1
    assert isinstance(nn_vstages["staging"][0], VStage)
    check_obj(
        nn_vstages["staging"][0].dict_state(),
        dict(
            artifact="nn",
            version="v0.0.1",
            stage="staging",
            is_active=True,
            ref="nn#staging#1",
            author=author,
            author_email=author_email,
            commit_hexsha=first_commit,
        ),
        skip_keys=skip_keys_assignment,
    )

    rf_artifact = artifacts["rf"]
    assert isinstance(rf_artifact, Artifact)
    assert rf_artifact.artifact == "rf"

    assert len(rf_artifact.versions) == 2
    assert all(isinstance(v, Version) for v in rf_artifact.versions)
    rf_ver1, rf_ver2 = rf_artifact.versions
    check_obj(
        rf_ver1.dict_state(),
        dict(
            artifact="rf",
            version="v1.2.3",
            is_active=True,
            ref="rf@v1.2.3",
            author=author,
            author_email=author_email,
            commit_hexsha=first_commit,
            discovered=False,
        ),
        skip_keys=skip_keys_registration,
    )
    check_obj(
        rf_ver2.dict_state(),
        dict(
            artifact="rf",
            version="v1.2.4",
            is_active=True,
            ref="rf@v1.2.4",
            author=author,
            author_email=author_email,
            commit_hexsha=second_commit,
            discovered=False,
        ),
        skip_keys=skip_keys_registration,
    )

    assert len(rf_artifact.get_events()) == 8
    assert all(
        isinstance(p, (Assignment, Unassignment, Registration, Deregistration, Commit))
        for p in rf_artifact.get_events()
    )
    assert all(
        isinstance(p, (Registration, Deregistration))
        for p in rf_ver1.get_events(indirect=False) + rf_ver2.get_events(indirect=False)
    )
    assert all(
        isinstance(p, (Assignment, Unassignment, Commit))
        for p in rf_ver1.get_events(direct=False) + rf_ver2.get_events(direct=False)
    )
    _rf_a4, rf_a1, _rf_c1 = rf_ver1.get_events(direct=False)
    rf_a3, rf_a2, _rf_c2 = rf_ver2.get_events(direct=False)

    check_obj(
        rf_a1.dict_state(),
        dict(
            artifact="rf",
            version="v1.2.3",
            stage="production",
            tag="rf#production#1",
            author=author,
            author_email=author_email,
            commit_hexsha=first_commit,
        ),
        skip_keys=skip_keys_assignment,
    )
    check_obj(
        rf_a3.dict_state(),
        dict(
            artifact="rf",
            version="v1.2.4",
            stage="production",
            tag="rf#production#3",
            author=author,
            author_email=author_email,
            commit_hexsha=second_commit,
        ),
        skip_keys=skip_keys_assignment,
    )
    check_obj(
        rf_a2.dict_state(),
        dict(
            artifact="rf",
            version="v1.2.4",
            stage="staging",
            tag="rf#staging#2",
            author=author,
            author_email=author_email,
            commit_hexsha=second_commit,
        ),
        skip_keys=skip_keys_assignment,
    )



import os
import re
from datetime import datetime
from enum import Enum
from typing import FrozenSet, Iterable, Optional, Union

from pydantic import BaseModel
from scmrepo.exceptions import RevError
from scmrepo.git import Git, GitTag

from .base import (
    Artifact,
    Assignment,
    BaseManager,
    BaseRegistryState,
    Deprecation,
    Deregistration,
    Registration,
    Unassignment,
)
from .constants import (
    ACTION,
    COUNTER,
    NAME,
    STAGE,
    TAG,
    VERSION,
    Action,
    name_to_tag,
    tag_re,
    tag_to_name,
)
from .exceptions import (
    InvalidTagName,
    MissingArg,
    NotImplementedInGTO,
    RefNotFound,
    TagExists,
    TagNotFound,
    UnknownAction,
)

COUNT_DELIMITER = "#"

TagTemplates = {
    # Action.CREATE: "{artifact}@",
    Action.DEPRECATE: "{artifact}@deprecated",
    Action.REGISTER: "{artifact}@{version}",
    Action.DEREGISTER: "{artifact}@{version}!",
    Action.ASSIGN: "{artifact}#{stage}",
    Action.UNASSIGN: "{artifact}#{stage}!",
}


def name_tag(
    action: Action,
    artifact: str,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    scm: Optional[Git] = None,
    simple: bool = False,
):
    if action not in TagTemplates:
        raise UnknownAction(action=action)

    tag = TagTemplates[action].format(
        artifact=name_to_tag(artifact), version=version, stage=stage
    )
    if simple:
        return tag
    if scm is None:
        raise MissingArg(arg="scm")
    counter = 0
    for t in scm.list_tags():
        parsed = parse_name(t, raise_on_fail=False)  # type: ignore
        if (
            parsed
            and parsed[NAME] == artifact
            and COUNTER in parsed
            and parsed[COUNTER] > counter
        ):
            counter = parsed[COUNTER]
    return f"{tag}{COUNT_DELIMITER}{counter+1}"


def parse_name(name: str, raise_on_fail: bool = True):

    match = re.search(tag_re, name)
    if raise_on_fail and not match:
        raise InvalidTagName(name)
    if match:
        parsed = {NAME: tag_to_name(match["artifact"])}
        if match["deprecated"]:
            parsed[ACTION] = Action.DEPRECATE
        if match[VERSION]:
            parsed[VERSION] = match[VERSION]
            parsed[ACTION] = (
                Action.DEREGISTER if match["cancel"] == "!" else Action.REGISTER
            )
        if match[STAGE]:
            parsed[STAGE] = match[STAGE]
            parsed[ACTION] = (
                Action.UNASSIGN if match["cancel"] == "!" else Action.ASSIGN
            )
        if match[COUNTER]:
            parsed[COUNTER] = int(match[COUNTER])
        return parsed
    return {}


class NAME_REFERENCE(Enum):
    TAG = TAG
    NAME = NAME


def parse_name_reference(name: str):
    parsed = parse_name(name, raise_on_fail=False)
    if not parsed:
        return NAME_REFERENCE.NAME, name
    return NAME_REFERENCE.TAG, parsed


class Tag(BaseModel):
    action: Action
    name: str
    version: Optional[str]
    stage: Optional[str]
    created_at: datetime
    tag: GitTag

    class Config:
        arbitrary_types_allowed = True


def parse_tag(tag: GitTag):
    return Tag(
        tag=tag,
        created_at=datetime.fromtimestamp(tag.tag_time),
        **parse_name(tag.name),
    )


def find(
    action: Union[Action, FrozenSet[Action]] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    scm: Optional[Git] = None,
    sort: str = "by_time",
    tags: Optional[Iterable[GitTag]] = None,
):
    if isinstance(action, Action):
        action = frozenset([action])
    if scm is None:
        raise MissingArg(arg="scm")
    result = []
    tag_names = [t.name for t in tags] if tags else scm.list_tags()
    for t in tag_names:
        try:
            parsed = parse_name(t)
        except InvalidTagName:
            continue
        if (  # pylint: disable=too-many-boolean-expressions
            parsed
            and (not action or parsed[ACTION] in action)
            and (not name or parsed.get(NAME) == name)
            and (not version or parsed.get(VERSION) == version)
            and (not stage or parsed.get(STAGE) == stage)
        ):
            tag = scm.get_tag(t)
            # remove lightweight tags
            if isinstance(tag, GitTag):
                result.append(tag)
    if sort == "by_time":
        return sorted(result, key=lambda t: t.tag_time)
    raise NotImplementedError(f"Unknown sort: {sort}")


def create_tag(  # pylint: disable=too-many-branches
    scm: Git,
    name: str,
    rev: str,
    message: str,
    tagger: str = None,
    tagger_email: str = None,
):
    try:
        rev = scm.resolve_rev(rev)
    except RevError as e:
        raise RefNotFound(ref=rev) from e
    if scm.get_tag(name):
        raise TagExists(name=name)

    if tagger:
        orig_name: Optional[str] = os.environ.get("GIT_COMMITTER_NAME")
        os.environ["GIT_COMMITTER_NAME"] = tagger
    else:
        orig_name = None
    if tagger_email:
        orig_email: Optional[str] = os.environ.get("GIT_COMMITTER_EMAIL")
        os.environ["GIT_COMMITTER_EMAIL"] = tagger_email
    else:
        orig_email = None
    try:
        scm.tag(name, target=rev, annotated=True, message=message)
    finally:
        if tagger:
            if orig_name:
                os.environ["GIT_COMMITTER_NAME"] = orig_name
            else:
                del os.environ["GIT_COMMITTER_NAME"]
        if tagger_email:
            if orig_email:
                os.environ["GIT_COMMITTER_EMAIL"] = orig_email
            else:
                del os.environ["GIT_COMMITTER_EMAIL"]


def delete_tag(scm: Git, name: str):
    ref = f"refs/tags/{name}"
    if not scm.get_ref(ref):
        raise TagNotFound(name=name)
    scm.remove_ref(ref)


def index_tag(artifact: Artifact, tag: GitTag) -> Artifact:
    event: Union[Deprecation, Registration, Deregistration, Assignment, Unassignment]
    mtag = parse_tag(tag)
    if mtag.action == Action.REGISTER:
        event = Registration(
            artifact=mtag.name,
            version=mtag.version,
            created_at=mtag.created_at,
            author=tag.tagger_name,
            author_email=tag.tagger_email,
            message=tag.message.strip(),
            commit_hexsha=tag.target,
            tag=tag.name,
        )
    elif mtag.action == Action.DEREGISTER:
        event = Deregistration(
            artifact=mtag.name,
            version=mtag.version,
            created_at=mtag.created_at,
            author=tag.tagger_name,
            author_email=tag.tagger_email,
            message=tag.message.strip(),
            commit_hexsha=tag.target,
            tag=tag.name,
        )
    elif mtag.action in (Action.ASSIGN, Action.UNASSIGN):
        version = artifact.find_version(
            commit_hexsha=tag.target, create_new=True
        ).version  # type: ignore
        if mtag.action == Action.ASSIGN:
            event = Assignment(
                artifact=mtag.name,
                version=version,
                stage=mtag.stage,
                created_at=mtag.created_at,
                author=tag.tagger_name,
                author_email=tag.tagger_email,
                message=tag.message.strip(),
                commit_hexsha=tag.target,
                tag=tag.name,
            )
        else:
            event = Unassignment(
                artifact=mtag.name,
                version=version,
                stage=mtag.stage,
                created_at=mtag.created_at,
                author=tag.tagger_name,
                author_email=tag.tagger_email,
                message=tag.message.strip(),
                commit_hexsha=tag.target,
                tag=tag.name,
            )
    elif mtag.action == Action.DEPRECATE:
        event = Deprecation(
            artifact=mtag.name,
            created_at=mtag.created_at,
            author=tag.tagger_name,
            author_email=tag.tagger_email,
            message=tag.message.strip(),
            commit_hexsha=tag.target,
            tag=tag.name,
        )
    artifact.add_event(event)
    return artifact


class TagManager(BaseManager):  # pylint: disable=abstract-method
    def update_state(self, state: BaseRegistryState) -> BaseRegistryState:
        # tags are sorted and then indexed by timestamp
        # this is important to check that history is not broken
        for tag in find(scm=self.scm, action=self.actions):
            state.update_artifact(
                index_tag(
                    state.find_artifact(parse_name(tag.name)["name"], create_new=True),
                    tag,
                )
            )
        return state


class TagArtifactManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.CREATE, Action.DEPRECATE))

    def create(self):  # pylint: disable=no-self-use
        raise NotImplementedInGTO(
            "If you want to create artifact, register a version or assign a stage for it"
        )

    def deprecate(
        self,
        name,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.DEPRECATE,
            name,
            scm=self.scm,
            simple=simple,
        )
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag


class TagVersionManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.REGISTER, Action.DEREGISTER))

    def register(
        self,
        name,
        version,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.REGISTER, name, version=version, scm=self.scm, simple=simple
        )
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag

    def deregister(
        self,
        name,
        version,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ):
        tag = name_tag(
            Action.DEREGISTER, name, version=version, scm=self.scm, simple=simple
        )
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag


class TagStageManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.ASSIGN, Action.UNASSIGN))

    def assign(
        self,
        name,
        stage,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> str:
        tag = name_tag(Action.ASSIGN, name, stage=stage, scm=self.scm, simple=simple)
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag

    def unassign(
        self,
        name,
        stage,
        rev,
        message,
        simple,
        author: Optional[str] = None,
        author_email: Optional[str] = None,
    ) -> str:
        tag = name_tag(Action.UNASSIGN, name, stage=stage, scm=self.scm, simple=simple)
        create_tag(
            self.scm,
            tag,
            rev=rev,
            message=message,
            tagger=author,
            tagger_email=author_email,
        )
        return tag



from pathlib import Path

from setuptools import find_packages, setup

install_requires = [
    "scmrepo>=1.3.0,<2",
    "typer>=0.4.1",
    "rich",
    "pydantic>=1.9.0,<2",
    "ruamel.yaml",
    "semver>=3.0.0",
    "entrypoints",
    "tabulate>=0.8.10",
    "funcy",
]


tests = [
    "pytest",
    "pytest-cov",
    "pytest-lazy-fixture==0.6.3",
    "pytest-mock",
    "pytest-test-utils",
    "pylint<2.14",
    # we use this to suppress pytest-related false positives in our tests.
    "pylint-pytest",
    # we use this to suppress some messages in tests, eg: foo/bar naming,
    # and, protected method calls in our tests
    "pylint-plugin-utils",
    "freezegun",
    "types-freezegun",
]


setup_args = dict(  # noqa: C408
    name="gto",
    use_scm_version=True,
    setup_requires=["setuptools_scm", "fastentrypoints>=0.12"],
    description="Version and deploy your models following GitOps principles",
    long_description=(Path(__file__).parent / "README.md").read_text(encoding="utf8"),
    long_description_content_type="text/markdown",
    author="Alexander Guschin",
    author_email="aguschin@iterative.ai",
    download_url="https://github.com/iterative/gto",
    license="Apache License 2.0",
    install_requires=install_requires,
    extras_require={"tests": tests},
    keywords="git repo repository artifact registry developer-tools collaboration",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    entry_points={
        "console_scripts": ["gto = gto.cli:app"],
        "gto.enrichment": [
            # "mlem = gto.ext_mlem:MlemEnrichment",
            # "dvc = gto.ext_dvc:DVCEnrichment",
            # "cli = gto.ext:CLIEnrichment",
            "gto = gto.index:GTOEnrichment",
        ],
    },
    zip_safe=False,
)

if __name__ == "__main__":
    setup(**setup_args)



from datetime import datetime
from typing import Any, Dict, FrozenSet, List, Optional, Sequence, Union

from pydantic import BaseModel
from scmrepo.git import Git

from gto.config import RegistryConfig
from gto.constants import (
    ASSIGNMENTS_PER_VERSION,
    VERSIONS_PER_STAGE,
    Action,
    VersionSort,
)
from gto.versions import SemVer

from .exceptions import (
    ArtifactNotFound,
    ManyVersions,
    NoStageForVersion,
    NotImplementedInGTO,
    VersionRequired,
    WrongArgs,
)


# EVENTS: deprecation, registration, deregistration, assignment, unassignment
class BaseEvent(BaseModel):
    priority: int
    addition: bool
    # some fields here are implementation details of tag-based
    # if we add other approaches beside tags, we'll need to fix this
    artifact: str
    created_at: datetime
    author: str
    author_email: str
    message: str
    commit_hexsha: str

    @property
    def event(self):
        return self.__class__.__name__.lower()

    def dict_state(self, exclude=None):
        state = self.dict(exclude=exclude)
        state["event"] = self.event
        return state

    @property
    def ref(self):
        return getattr(self, "tag", self.commit_hexsha)


class Commit(BaseEvent):
    priority = 0
    addition = True
    version: str
    enrichments: List[Any]
    committer: str
    committer_email: str

    def __str__(self):
        return f'Artifact "{self.artifact}" is annotated'


class Creation(BaseEvent):
    priority = 1
    addition = True
    tag: str

    def __str__(self):
        return f'Artifact "{self.artifact}" was created'


class Deprecation(BaseEvent):
    priority = 2
    addition = False
    tag: str

    def __str__(self):
        return f'Artifact "{self.artifact}" was deprecated'


class Registration(BaseEvent):
    priority = 3
    addition = True
    tag: str
    version: str

    def __str__(self):
        return f'Version "{self.version}" of artifact "{self.artifact}" was registered'


class Deregistration(BaseEvent):
    priority = 4
    addition = False
    tag: str
    version: str

    def __str__(self):
        return (
            f'Version "{self.version}" of artifact "{self.artifact}" was deregistered'
        )


class Assignment(BaseEvent):
    priority = 5
    addition = True
    tag: str
    version: str
    stage: str

    def __str__(self) -> str:
        return f'Stage "{self.stage}" was assigned to version "{self.version}" of artifact "{self.artifact}"'


class Unassignment(BaseEvent):
    priority = 6
    addition = False
    tag: str
    version: str
    stage: str

    def __str__(self) -> str:
        return f'Stage "{self.stage}" was unassigned from version "{self.version}" of artifact "{self.artifact}"'


# ENTITIES: Artifact, Version, Stage
class BaseObject(BaseModel):
    artifact: str

    def add_event(self, event: BaseEvent):
        raise NotImplementedError()

    def get_events(
        self, direct=True, indirect=True, ascending: bool = False
    ) -> Sequence[BaseEvent]:
        raise NotImplementedError()

    @property
    def is_active(self):
        raise NotImplementedError()

    @property
    def activated_at(self):
        raise NotImplementedError()

    @property
    def authoring_event(self):
        addition_events = [
            e for e in self.get_events(direct=True, indirect=False) if e.addition
        ]
        if addition_events:
            return addition_events[0]
        events = self.get_events(direct=False, indirect=True)
        except_enrichment = [e for e in events if e.event != "enrichment"]
        if except_enrichment:
            return except_enrichment[0]
        return events[0]

    @property
    def created_at(self):
        return self.authoring_event.created_at

    @property
    def author(self):
        return self.authoring_event.author

    @property
    def author_email(self):
        return self.authoring_event.author_email

    @property
    def message(self):
        return self.authoring_event.message

    @property
    def ref(self):
        return self.authoring_event.ref

    def dict_state(self, exclude=None):
        version = self.dict(exclude=exclude)
        version["is_active"] = self.is_active
        version["activated_at"] = self.activated_at
        version["created_at"] = self.created_at
        version["author"] = self.author
        version["author_email"] = self.author_email
        version["message"] = self.message
        version["ref"] = self.ref
        return version


class VStage(BaseObject):
    commit_hexsha: str
    version: str
    stage: str
    assignments: List[Assignment] = []
    unassignments: List[Unassignment] = []

    def add_event(self, event: BaseEvent):
        if event in self.get_events():
            return event
        if isinstance(event, Assignment):
            self.assignments.append(event)
            self.assignments.sort(key=lambda e: e.created_at)
        elif isinstance(event, Unassignment):
            self.unassignments.append(event)
            self.unassignments.sort(key=lambda e: e.created_at)
        else:
            raise NotImplementedInGTO(f"Unknown event {event} of class {type(event)}")
        return event

    def get_events(
        self, direct=True, indirect=True, ascending=False
    ) -> Sequence[BaseEvent]:  # pylint: disable=unused-argument
        return sorted(
            self.assignments + self.unassignments if direct else [], key=lambda e: e.created_at  # type: ignore
        )[:: 1 if ascending else -1]

    @property
    def is_active(self):
        return isinstance(self.get_events()[0], Assignment)

    @property
    def activated_at(self):
        if isinstance(self.get_events()[0], Assignment):
            return self.get_events()[0].created_at
        return None

    def dict_state(self, exclude=None):
        state = super().dict_state(exclude=exclude)
        state["assignments"] = [a.dict_state() for a in self.assignments]
        state["unassignments"] = [a.dict_state() for a in self.unassignments]
        return state


class Version(BaseObject):
    commit_hexsha: str
    version: str
    enrichments: List[Commit] = []
    registrations: List[Registration] = []
    deregistrations: List[Deregistration] = []
    stages: Dict[str, VStage] = {}

    def add_event(self, event: BaseEvent):
        if event in self.get_events():
            return event
        if isinstance(event, Registration):
            self.registrations.append(event)
            self.registrations.sort(key=lambda e: e.created_at)
        elif isinstance(event, Deregistration):
            self.deregistrations.append(event)
            self.deregistrations.sort(key=lambda e: e.created_at)
        elif isinstance(event, Commit):
            self.enrichments.append(event)
            self.enrichments.sort(key=lambda e: e.created_at)
        elif isinstance(event, (Assignment, Unassignment)):
            self.get_vstage(event.stage, create_new=True).add_event(event)
        else:
            raise NotImplementedInGTO(f"Unknown event {event} of class {type(event)}")
        return event

    def get_events(self, direct=True, indirect=True, ascending=False):
        return sorted(
            (self.registrations + self.deregistrations if direct else [])
            + (
                self.enrichments
                + [e for s in self.stages.values() for e in s.get_events()]
                if indirect
                else []
            ),
            key=lambda e: e.created_at,
        )[:: 1 if ascending else -1]

    @property
    def is_active(self):
        direct_events = self.get_events(direct=True, indirect=False)
        if direct_events:
            return isinstance(direct_events[0], Registration)
        indirect_events = self.get_events(direct=False, indirect=True)
        if indirect_events:
            return True
        return False

    @property
    def activated_at(self):
        # TODO: handle the case with deregistration?
        if self.registrations:
            return self.registrations[0].created_at
        return self.get_events(ascending=True)[0].created_at

    @property
    def is_registered(self):
        """Tells if this is an explicitly registered version"""
        return len(self.registrations) > 0  # SemVer.is_valid(self.name)

    @property
    def discovered(self):
        return len(self.get_events(direct=True, indirect=False)) == 0

    @property
    def get_enrichments_info(self):
        if len(self.enrichments) > 1:
            raise NotImplementedInGTO(
                "Multiple enrichments for a single version are not supported"
            )
        return self.enrichments[0].enrichments

    @property
    def semver(self):
        return SemVer(self.version)

    def get_vstage(self, stage, create_new=False):
        if create_new and stage not in self.stages:
            self.stages[stage] = VStage(
                artifact=self.artifact,
                version=self.version,
                stage=stage,
                commit_hexsha=self.commit_hexsha,
            )
        if stage in self.stages:
            return self.stages[stage]
        raise NoStageForVersion(self.artifact, self.version, stage)

    def get_vstages(self, active_only=True, ascending=False):
        return sorted(
            [s for s in self.stages.values() if not active_only or s.is_active],
            key=lambda s: s.activated_at,
        )[:: 1 if ascending else -1]

    def dict_state(self, exclude=None, assignments_per_version=ASSIGNMENTS_PER_VERSION):
        if assignments_per_version < -1:
            raise WrongArgs("'assignments_per_version' must be >= -1")
        version = super().dict_state(exclude=exclude)
        version["discovered"] = self.discovered
        version["stages"] = [stage.dict_state() for stage in self.get_vstages()]
        if assignments_per_version >= 0:
            version["stages"] = version["stages"][:assignments_per_version]
        return version


def sort_versions(
    versions,
    sort=VersionSort.SemVer,
    ascending=False,
    version="version",
    timestamp="created_at",
):
    """This function is used both in Studio and in GTO"""

    def get(obj, key):
        if isinstance(obj, dict):
            return obj[key]
        if isinstance(obj, BaseModel):
            return getattr(obj, key)
        raise NotImplementedError("Can sort either dict or BaseModel")

    sort = sort if isinstance(sort, VersionSort) else VersionSort[sort]
    if sort == VersionSort.SemVer:
        # sorting SemVer versions in a right way
        sorted_versions = sorted(
            (v for v in versions if SemVer.is_valid(get(v, version))),
            key=lambda x: SemVer(get(x, version)),
        )[:: 1 if ascending else -1]
        # sorting hexsha versions alphabetically
        sorted_versions.extend(
            sorted(
                (v for v in versions if not SemVer.is_valid(get(v, version))),
                key=lambda x: get(x, version),
            )[:: 1 if ascending else -1]
        )
    else:
        sorted_versions = sorted(
            versions,
            key=lambda x: get(x, timestamp),
        )[:: 1 if ascending else -1]
    return sorted_versions


class Artifact(BaseObject):
    versions: List[Version]
    creations: List[Creation] = []
    deprecations: List[Deprecation] = []

    def add_event(self, event: BaseEvent):
        if event in self.get_events():
            return event
        if isinstance(event, Creation):
            self.creations.append(event)
            self.creations.sort(key=lambda e: e.created_at)
        elif isinstance(event, Deprecation):
            self.deprecations.append(event)
            self.deprecations.sort(key=lambda e: e.created_at)
        elif isinstance(
            event, (Registration, Deregistration, Assignment, Unassignment, Commit)
        ):
            self.find_version(
                event.version, commit_hexsha=event.commit_hexsha, create_new=True
            ).add_event(  # type: ignore
                event
            )
        else:
            raise NotImplementedInGTO(f"Unknown event {event} of class {type(event)}")
        return event

    def get_events(
        self, direct=True, indirect=True, ascending=False
    ) -> Sequence[BaseEvent]:
        return sorted(
            (self.creations + self.deprecations if direct else [])  # type: ignore
            + ([e for v in self.versions for e in v.get_events()] if indirect else []),
            key=lambda e: e.created_at,
        )[:: 1 if ascending else -1]

    @property
    def is_active(self):
        if len(self.get_events()) == 0:
            return False
        return not isinstance(self.get_events()[0], Deprecation)

    @property
    def activated_at(self):
        # TODO: handle the case with deprecation
        if self.creations:
            return self.creations[0].created_at
        return self.get_events(ascending=True)[0].created_at

    @property
    def is_registered(self):
        """Tells if this is an a registered artifact - i.e. there are Git tags for it"""
        return not all(
            isinstance(e, Commit) for e in self.get_events(direct=True, indirect=True)
        )

    @property
    def unique_stages(self):
        return set(self.get_vstages())

    def __repr__(self) -> str:
        versions = ", ".join(f"'{v.version}'" for v in self.versions)
        stages = ", ".join(f"'{p}'" for p in self.unique_stages)
        return f"Artifact(versions=[{versions}], stages=[{stages}])"

    def get_versions(
        self,
        active_only=True,
        include_non_explicit=False,
        include_discovered=False,
        sort=VersionSort.SemVer,
        ascending=False,
    ) -> List[Version]:
        versions = [
            v
            for v in self.versions
            if not active_only
            or v.is_active
            and (
                (v.is_registered and not v.discovered)
                or (include_discovered and v.discovered)
                or (include_non_explicit and not v.is_registered)
            )
        ]
        return sort_versions(versions, sort=sort, ascending=ascending)

    def get_latest_version(
        self, registered_only=False, sort=VersionSort.SemVer
    ) -> Optional[Version]:
        versions = self.get_versions(
            include_non_explicit=not registered_only, sort=sort
        )
        if versions:
            return versions[0]
        return None

    def get_vstages(
        self,
        registered_only=False,
        assignments_per_version=ASSIGNMENTS_PER_VERSION,
        versions_per_stage=VERSIONS_PER_STAGE,
        sort=VersionSort.SemVer,
    ):
        if assignments_per_version < -1:
            raise WrongArgs("'assignments_per_version' must be >= -1")
        if versions_per_stage < -1:
            raise WrongArgs("'versions_per_stage' must be >=-1")
        versions = self.get_versions(
            include_non_explicit=not registered_only, sort=sort
        )
        stages: Dict[str, List[VStage]] = {}
        assignments = [
            a
            for version in versions
            for a in (
                version.get_vstages(ascending=True)[:assignments_per_version]
                if assignments_per_version > -1
                else version.get_vstages(ascending=True)
            )
        ]
        if sort == VersionSort.Timestamp:
            assignments = sorted(assignments, key=lambda a: a.created_at)[::-1]
        for a in assignments:
            if a.stage not in stages:
                stages[a.stage] = []
            if (
                versions_per_stage > -1  # pylint: disable=chained-comparison
                and len(stages[a.stage]) >= versions_per_stage
            ):
                continue
            if a.version not in [i.version for i in stages[a.stage]]:
                stages[a.stage].append(a)
        return stages

    @property
    def discovered(self):
        return any(not v.discovered for v in self.versions)

    def find_version(
        self,
        name: str = None,
        commit_hexsha: str = None,
        raise_if_not_found: bool = False,
        allow_multiple=False,
        include_discovered=False,
        create_new=False,
    ) -> Union[None, Version, List[Version]]:
        versions = [
            v
            for v in self.versions
            if (v.version == name if name else True)
            and (v.commit_hexsha == commit_hexsha if commit_hexsha else True)
            and (True if include_discovered else not v.discovered)
        ]
        if create_new and not versions:
            v = Version(
                artifact=self.artifact,
                version=name or commit_hexsha,
                commit_hexsha=commit_hexsha,
            )
            self.versions.append(v)
            versions = [v]
        if allow_multiple and versions:
            return versions
        if raise_if_not_found and not versions:
            raise VersionRequired(name=self.artifact)
        if len(versions) > 1:
            raise ManyVersions(
                name=self.artifact,
                versions=[v.version for v in versions],
            )
        return versions[0] if versions else None

    def find_version_at_commit(
        self, commit_hexsha: str, latest_datetime: datetime = None
    ):
        return [
            v
            for v in self.find_version(  # type: ignore
                commit_hexsha=commit_hexsha,
                raise_if_not_found=True,
                allow_multiple=True,
            )
            if (v.activated_at <= latest_datetime if latest_datetime else True)  # type: ignore
        ][-1]


class BaseRegistryState(BaseModel):
    artifacts: Dict[str, Artifact] = {}

    class Config:
        arbitrary_types_allowed = True

    def add_artifact(self, name):
        self.artifacts[name] = Artifact(artifact=name, versions=[])

    def update_artifact(self, artifact: Artifact):
        self.artifacts[artifact.artifact] = artifact

    def get_artifacts(self):
        return self.artifacts

    def find_artifact(self, name: str, create_new=False) -> Artifact:
        if name not in self.artifacts:
            if create_new:
                self.artifacts[name] = Artifact(artifact=name, versions=[])
            else:
                raise ArtifactNotFound(name)
        return self.artifacts[name]

    @property
    def unique_stages(self):
        return sorted({p for o in self.artifacts.values() for p in o.unique_stages})

    def find_commit(self, name, version):
        return (
            self.find_artifact(name)
            .find_version(name=version, raise_if_not_found=True)
            .commit_hexsha
        )

    def which(
        self,
        name,
        stage,
        raise_if_not_found=True,
        assignments_per_version=None,
        versions_per_stage=None,
        registered_only=False,
    ):
        """Return stage active in specific stage"""
        assigned = self.find_artifact(name).get_vstages(
            registered_only=registered_only,
            assignments_per_version=assignments_per_version,
            versions_per_stage=versions_per_stage,
        )
        if stage in assigned:
            return assigned[stage]
        if raise_if_not_found:
            raise ValueError(f"Stage {stage} not found for {name}")
        return None


class BaseManager(BaseModel):
    scm: Git
    actions: FrozenSet[Action]
    config: RegistryConfig

    class Config:
        arbitrary_types_allowed = True

    def update_state(
        self, state: BaseRegistryState
    ) -> BaseRegistryState:  # pylint: disable=no-self-use
        raise NotImplementedError
